#include <dlfcn.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>

#include "adl_sdk.h"

#define ARRAY_SIZE(x) (int)(sizeof(x) / sizeof(x[0]))

#define LM63_SLAVE_ADDRESS              0x4C

#define LM63_REG_LOCAL_TEMP             0x00
#define LM63_REG_LOCAL_HIGH             0x05

#define LM63_REG_REMOTE_TEMP_MSB        0x01
#define LM63_REG_REMOTE_TEMP_LSB        0x10
#define LM63_REG_REMOTE_OFFSET_MSB      0x11
#define LM63_REG_REMOTE_OFFSET_LSB      0x12


typedef int (*ADL_MAIN_CONTROL_CREATE)(ADL_MAIN_MALLOC_CALLBACK, int);
typedef int (*ADL_MAIN_CONTROL_DESTROY)();
typedef int (*ADL_ADAPTER_NUMBEROFADAPTERS_GET)(int *);
typedef int (*ADL_ADAPTER_ADAPTERINFO_GET)(LPAdapterInfo, int);
typedef int (*ADL_DISPLAY_WRITEANDREADI2CREV_GET)(int, int *, int *);
typedef int (*ADL_DISPLAY_WRITEANDREADI2C)(int, ADLI2C *);

ADL_MAIN_CONTROL_CREATE ADL_Main_Control_Create;
ADL_MAIN_CONTROL_DESTROY ADL_Main_Control_Destroy;
ADL_ADAPTER_NUMBEROFADAPTERS_GET ADL_Adapter_NumberOfAdapters_Get;
ADL_ADAPTER_ADAPTERINFO_GET ADL_Adapter_AdapterInfo_Get;
ADL_DISPLAY_WRITEANDREADI2CREV_GET ADL_Display_WriteAndReadI2CRev_Get;
ADL_DISPLAY_WRITEANDREADI2C ADL_Display_WriteAndReadI2C;

static char *adl_result_ok[] = {
  "OK",
  "OK (with warning)",
  "OK (need mode change)",
  "OK (need restart)",
  "OK (need to wait)",
};

static char *adl_result_error[] = {
  "OK",
  "Generic error",
  "ADL not initialized",
  "Invalid parameter",
  "Invalid parameter size",
  "Invalid ADL index",
  "Invalid controller index",
  "Invalid display index",
  "Function not supported",
  "Null pointer",
  "Adapter disabled",
  "Invalid callback",
  "Display resource conflict",
};
static char adl_result_unknown[0x40];

static void * __stdcall ADL_Main_Memory_Alloc(int size) {
  void *buffer = malloc(size);
  return buffer;
}

static void __stdcall ADL_Main_Memory_Free(void **buffer) {
  if(*buffer != NULL) {
    free(*buffer);
    *buffer = NULL;
  }
}

static char * adl_error(int error) {
  memset(adl_result_unknown, 0, sizeof(adl_result_unknown));

  if(error < 0) {
    if((-error) >= ARRAY_SIZE(adl_result_error)) {
      sprintf(adl_result_unknown, "Unknown error status %d", error);
      return adl_result_unknown;
    } else {
      return adl_result_error[-error];
    }
  } else {
    if(error >= ARRAY_SIZE(adl_result_ok)) {
      sprintf(adl_result_unknown, "Unkown OK status %d", error);
      return adl_result_unknown;
    } else {
      return adl_result_ok[error];
    }
  }
}

static void usage(void) {
  fprintf(stdout, "\n");
  fprintf(stdout, "Usage: adli2c [-a adapterIndex] [-t thermalControllerIndex] [-h]\n");
  fprintf(stdout, "  -a adapterIndex  index of the adapter to use (default 0)\n");
  fprintf(stdout, "  -h               display help\n");
  fprintf(stdout, "\n");
}

#if defined(DEBUG)
static void dump_adapter_info(LPAdapterInfo adapterInfo, int numAdapters) {
  int i;

  for(i = 0; i < numAdapters; i++) {
	fprintf(stdout, "--- Adaptor %d/%d ---\n", i + 1, numAdapters);
	fprintf(stdout, "\tiSize %d\n", adapterInfo[i].iSize);
	fprintf(stdout, "\tiAdapterIndex %d\n", adapterInfo[i].iAdapterIndex);
	fprintf(stdout, "\tstrUDID %s\n", adapterInfo[i].strUDID);
	fprintf(stdout, "\tiBusNumber %d\n", adapterInfo[i].iBusNumber);
	fprintf(stdout, "\tiDeviceNumber %d\n", adapterInfo[i].iDeviceNumber);
	fprintf(stdout, "\tiFunctionNumber %d\n", adapterInfo[i].iFunctionNumber);
	fprintf(stdout, "\tiVendorID 0x%x\n", adapterInfo[i].iVendorID);
	fprintf(stdout, "\tstrAdapterName %s\n", adapterInfo[i].strAdapterName);
	fprintf(stdout, "\tstrDisplayName %s\n", adapterInfo[i].strDisplayName);
	fprintf(stdout, "\tiPresent %d\n", adapterInfo[i].iPresent);
  }
}
#endif

static int adl_i2c_read_byte(int adapterIndex, int line, int address, int offset, char *value) {
  int ret;
  ADLI2C i2c;
  char i2cData[ADL_DL_I2C_MAXDATASIZE];

  if(value == NULL) {
	return ADL_ERR;
  }

  memset(&i2c, 0, sizeof(ADLI2C));
  i2c.iSize = sizeof(ADLI2C);
  i2c.iLine = line;
  i2c.iAddress = address << 1;
  i2c.iOffset = offset;
  i2c.iAction = ADL_DL_I2C_ACTIONREAD;
  i2c.iSpeed = 1;
  i2c.iDataSize = 1;
  i2c.pcData = i2cData;
  memset(&i2cData, 0, ADL_DL_I2C_MAXDATASIZE);
  
  ret = ADL_Display_WriteAndReadI2C(adapterIndex, &i2c);
  if(ret == ADL_OK) {
    *value = i2cData[0];
  }

  return ret;
}

int main(int argc, char **argv) {
  int opt;
  int err;
  int ret;
  void *handle;

  int numAdapters;
  int adapterIndex;
  LPAdapterInfo adapterInfo;
  
  int i2cMajor;
  int i2cMinor;
  int line;
  char tmp;
  uint16_t remoteTemp;

  err = EXIT_FAILURE;

  adapterIndex = 0;
  while((opt = getopt(argc, argv, "a:h")) != -1) {
    switch(opt) {
    case 'a':
      sscanf(optarg, "%d", &adapterIndex);
      break;
    case 'h':
      usage();
      return EXIT_SUCCESS;
    default:
      fprintf(stderr, "Invalid option %c\n", opt);
      usage();
      return err;
    }
  }
  
  handle = dlopen("libatiadlxx.so", RTLD_LAZY | RTLD_GLOBAL);
  if(handle == NULL) {
    fprintf(stderr, "Failed to open AMD library (%s)\n", strerror(errno));
    return err;
  }
  
  ADL_Main_Control_Create = 
    (ADL_MAIN_CONTROL_CREATE)dlsym(handle, "ADL_Main_Control_Create");
  ADL_Main_Control_Destroy =
    (ADL_MAIN_CONTROL_DESTROY)dlsym(handle, "ADL_Main_Control_Destroy");
  ADL_Adapter_NumberOfAdapters_Get =
    (ADL_ADAPTER_NUMBEROFADAPTERS_GET)dlsym(handle, "ADL_Adapter_NumberOfAdapters_Get");
  ADL_Adapter_AdapterInfo_Get =
    (ADL_ADAPTER_ADAPTERINFO_GET)dlsym(handle, "ADL_Adapter_AdapterInfo_Get");
  ADL_Display_WriteAndReadI2CRev_Get =
    (ADL_DISPLAY_WRITEANDREADI2CREV_GET)dlsym(handle, "ADL_Display_WriteAndReadI2CRev_Get");
  ADL_Display_WriteAndReadI2C =
    (ADL_DISPLAY_WRITEANDREADI2C)dlsym(handle, "ADL_Display_WriteAndReadI2C");

  if((ADL_Main_Control_Create == NULL) ||
     (ADL_Main_Control_Destroy == NULL) ||
     (ADL_Adapter_NumberOfAdapters_Get == NULL) ||
     (ADL_Adapter_AdapterInfo_Get == NULL) ||
     (ADL_Display_WriteAndReadI2CRev_Get == NULL) ||
     (ADL_Display_WriteAndReadI2C == NULL)) {
    fprintf(stderr, "Failed to load symbols from library\n");
    goto err_dl;
  }
  
  ret = ADL_Main_Control_Create (ADL_Main_Memory_Alloc, 1);
  if(ret != ADL_OK) {
    fprintf(stderr, "Failed to initialize (%s)\n", adl_error(ret));
    goto err_dl;
  }

  ret = ADL_Adapter_NumberOfAdapters_Get(&numAdapters);
  if(ret != ADL_OK) {
    fprintf(stderr, "Failed to retrieve number of adapters (%s)\n", adl_error(ret));
    goto err_close;
  }
  
  if(numAdapters == 0) {
    fprintf(stderr, "No adapter found\n");
    goto err_close;
  } else {
    fprintf(stdout, "Found %d adapters\n", numAdapters);
  }
  
  if((adapterIndex < 0) || (adapterIndex >= numAdapters)) {
    fprintf(stderr, "Bad adapter index\n");
    goto err_close;
  } else {
    fprintf(stdout, "Using adapter index %d\n", adapterIndex);
  }
  
  adapterInfo = (LPAdapterInfo)malloc(sizeof(AdapterInfo) * numAdapters);
  if(adapterInfo == NULL) {
    fprintf(stderr, "Failed to allocate space for adapters informations\n");
    goto err_close;
  }
  memset(adapterInfo, 0, sizeof(AdapterInfo) * numAdapters);
  
  ret = ADL_Adapter_AdapterInfo_Get(adapterInfo, sizeof(AdapterInfo) * numAdapters);
  if(ret != ADL_OK) {
    fprintf(stderr, "Failed to retrieve adapter info (%s)\n", adl_error(ret));
    goto err_free_adapter;
  } else {
#if defined (DEBUG)
    dump_adapter_info(adapterInfo, numAdapters);
#endif
  }
  
  i2cMajor = 0;
  i2cMinor = 0;
  ret = ADL_Display_WriteAndReadI2CRev_Get(adapterIndex, &i2cMajor, &i2cMinor);
  if(ret != ADL_OK) {
    fprintf(stderr, "Failed to retrieve I2C API revision (%s)\n", adl_error(ret));
    goto err_free_adapter;
  } else {
#if defined(DEBUG)
    fprintf(stdout, "Adaptor is using I2C API revision %d.%d\n", i2cMajor, i2cMinor);
#endif
  }
  
  /* Read local temp */
  /* Which line to use ???
   * unable to find the matching between OEM, OD_CONTROL and OEM2 lines
   * and evaluation board schematics...
   */
  line = ADL_DL_I2C_LINE_OEM;
  ret = adl_i2c_read_byte(adapterIndex, line, 
			  LM63_SLAVE_ADDRESS, LM63_REG_LOCAL_TEMP, &tmp);
  if(ret != ADL_OK) {
    fprintf(stderr, "Failed to read LM63_REG_LOCAL_TEMP register (%s)\n", adl_error(ret));
  } else {
    fprintf(stdout, "Local temperature %d\n", tmp);
  }
  ret = adl_i2c_read_byte(adapterIndex, line, 
			  LM63_SLAVE_ADDRESS, LM63_REG_LOCAL_HIGH, &tmp);
  if(ret != ADL_OK) {
    fprintf(stderr, "Failed to read LM63_REG_LOCAL_HIGH register (%s)\n", adl_error(ret));
  } else {
    fprintf(stdout, "Local high setpoint %d\n", tmp);
  }
  
  /* Read remote temp */
  ret = adl_i2c_read_byte(adapterIndex, line, 
			  LM63_SLAVE_ADDRESS, LM63_REG_REMOTE_TEMP_MSB, &tmp);
  if(ret != ADL_OK) {
    fprintf(stderr, "Failed to read LM63_REG_REMOTE_TEMP_MSB register (%s)\n", adl_error(ret));
  } else {
    remoteTemp = tmp << 8;
  }
  ret = adl_i2c_read_byte(adapterIndex, line, 
			  LM63_SLAVE_ADDRESS, LM63_REG_REMOTE_TEMP_LSB, &tmp);
  if(ret != ADL_OK) {
    fprintf(stderr, "Failed to read LM63_REG_REMOTE_TEMP_MSB register (%s)\n", adl_error(ret));
  } else {
    remoteTemp |= tmp;
    fprintf(stdout, "Remote temperature %d\n", remoteTemp / 32 * 125);
  }
  
  err = EXIT_SUCCESS;

err_free_adapter:
  ADL_Main_Memory_Free((void **)&adapterInfo);

err_close:
  ADL_Main_Control_Destroy();

err_dl:
  dlclose(handle);

  return err;
}
